declare module "windows.security.credentials" {
  export enum WebAccountState {
    none,
    connected,
    error,
  }

  export class PasswordCredential {
    userName: String;
    resource: String;
    password: String;
    properties: Object;
    constructor();
    constructor(resource: String, userName: String, password: String);

    retrievePassword(): void;

  }

  export class PasswordVault {
    constructor();

    add(credential: PasswordCredential): void;

    remove(credential: PasswordCredential): void;

    retrieve(resource: String, userName: String): PasswordCredential;

    findAllByResource(resource: String): Object;

    findAllByUserName(userName: String): Object;

    retrieveAll(): Object;

  }

  export class PasswordCredentialPropertyStore {
    constructor();

    lookup(key: String): Object;

    hasKey(key: String): Boolean;

    getView(): Object;

    insert(key: String, value: Object): Boolean;

    remove(key: String): void;

    clear(): void;

    first(): Object;

    addListener(type: "MapChanged", listener: (ev: Event) => void): void ;
    removeListener(type: "MapChanged", listener: (ev: Event) => void): void ;
    on(type: "MapChanged", listener: (ev: Event) => void): void ;
    off(type: "MapChanged", listener: (ev: Event) => void): void ;
    
    addListener(type: string, listener: (ev: Event) => void): void ;
    removeListener(type: string, listener: (ev: Event) => void): void ;
    on(type: string, listener: (ev: Event) => void): void ;
    off(type: string, listener: (ev: Event) => void): void ;
    

  }

  export class WebAccountProvider {
    displayName: String;
    iconUri: Object;
    id: String;
    constructor();
    constructor(id: String, displayName: String, iconUri: Object);

  }

  export class WebAccount {
    state: WebAccountState;
    userName: String;
    webAccountProvider: WebAccountProvider;
    constructor();
    constructor(webAccountProvider: WebAccountProvider, userName: String, state: WebAccountState);

  }

}



